# XPDFJ.pm - PDFJ XML interface
# 2002-3 <nakajima@netstock.co.jp>
package XPDFJ;
use PDFJ qw(UTF8);
use XML::Parser;
#use Data::Dumper;
use Safe;
use Carp;
use strict;
use vars qw($VERSION $bytesimport);

$VERSION = 0.11;

my $VarnamePtn = qr/^\$[A-Za-z_]\w*(\{[A-Za-z_]\w*\})?$/;

if( $] > 5.007 ) {
	require bytes;
	$bytesimport = \&bytes::import;
}

sub new {
	my($class, %args) = @_;
	my $safe = new Safe;
	$safe->permit_only(qw(:default));
	$safe->share(qw(istext addcomma autonl));
	$safe->share(qw($bytesimport)) if $] > 5.007;
	my $self = bless {safe => $safe}, $class;
	$args{dopath} ||= '';
	$args{VERSION} = $VERSION;
	for my $key(keys %args) {
		$self->var("\$XPDFJ{$key}", $args{$key});
	}
	$self;
}

sub verbose {
	my($self, $value) = @_;
	if( defined $value ) {
		$self->{verbose} = $value;
	}
	$self->{verbose};
}

sub parsefile { 
	my($self, $file, %args) = @_;
	print "parse '$file'\n" if $self->verbose >= 0;
	my $parser = new XML::Parser(Style => 'Tree');
	$self->{tree} = $parser->parsefile($file);
	$self->root_parse(\%args);
}

sub parse { 
	my($self, $data, %args) = @_;
	print "parse data in memory\n" if $self->verbose >= 0;
	my $parser = new XML::Parser(Style => 'Tree');
	$self->{tree} = $parser->parse($data);
	$self->root_parse(\%args);
}

sub root_parse {
	my($self, $args) = @_;
	my($element) = $self->elements($self->{tree});
	my($tag, $attributes, $children) = @$element;
	die "root must be 'XPDFJ'" if $tag ne 'XPDFJ';
	my %args = $self->args($attributes);
	my $version = delete $args{version};
	die "version '$version' required but my version is '$VERSION'"
		if $version && $version > $VERSION;
	%args = (%args, %$args) if ref($args) eq 'HASH';
	$self->method_parse($children, undef, \%args);
}

sub method_parse {
	my($self, $tree, $this, $args, $withtext) = @_;
	my @result;
	$self->enter_local($args) if $args;
	for my $elem($self->elements($tree)) {
		my($tag, $attributes, $children) = @$elem;
		if( $tag eq '0' ) {
			next unless $withtext;
			$children = $self->var($children) if $children =~ /$VarnamePtn/;
			if( $withtext eq 'autonl' ) {
				push @result, autonl($children);
			} else {
				push @result, $children;
			}
		} else {
			push @result, $self->call_method($tag, $attributes, $children, $this);
		}
	}
	$self->leave_local if $args;
	wantarray ? @result : @result <= 1 ? $result[0] : \@result;
}

sub call_method {
	my($self, $tag, $attributes, $children, $this) = @_;
	my @result;
	eval { # for error handling
	if( $self->{alias}{$tag} ) {
		my $aliasargs = $self->{aliasargs}{$tag};
		if( %$aliasargs ) {
			$attributes = {%$aliasargs, %$attributes};
		}
		$tag = $self->{alias}{$tag};
	}
	my $func = "do_$tag";
	if( $self->can($func) ) {
		@result = $self->$func($attributes, $children, $this);
		print "$func($attributes, $children, $this) => @result\n" 
			if $self->verbose > 1;
	} else {
		my $callchildren;
		($callchildren, $children) = $self->split_call($children);
		if( $self->{attributesname}{$tag} ) {
			$attributes = {$self->{attributesname}{$tag} =>
				{$self->args($attributes)}};
		}
		my %args;
		my $contentsmode = $self->{contentsmode}{$tag};
		my $contentsname = $self->{contentsname}{$tag} || 'contents';
		if( $contentsmode eq '' ) {
			%args = $self->args($attributes, $children);
		} elsif( $contentsmode eq 'method' ) {
			%args = ($self->args($attributes), 
			$contentsname => [$self->method_parse($children, $this)]);
		} elsif( $contentsmode eq 'text' ) {
			%args = ($self->args($attributes), 
			$contentsname => [$self->method_parse($children, $this, undef, 1)]);
		} elsif( $contentsmode eq 'autonl' ) {
			%args = ($self->args($attributes), 
			$contentsname => 
			[$self->method_parse($children, $this, undef, 'autonl')]);
		} elsif( $contentsmode eq 'raw' ) {
			%args = ($self->args($attributes), 
			$contentsname => $children);
		} else {
			die "unknown contentsmode '$contentsmode'";
		}
		if( $self->{defargs}{$tag} ) {
			my %defargs = %{$self->{defargs}{$tag}};
			$self->subst_vars(\%defargs);
			for my $key(keys %defargs) {
				$args{$key} = $defargs{$key} unless defined $args{$key};
			}
		}
		my $call = delete $args{call};
		my %callargs = $self->args(undef, $callchildren);
		push @$call, @{$callargs{call}} if $callargs{call};
		my $varname = delete $args{setvar};
		$varname ||= '$Doc' if $tag eq 'Doc';
		if( $self->{def}{$tag} ) {
			@result = $self->method_parse($self->{def}{$tag}, $this, \%args);
		} elsif( $tag =~ /^[A-Z]/ ) {
			no strict 'refs';
			print "$tag(@{[%args]})" if $self->verbose;
			@result = &$tag(\%args);
			print " => @result\n" if $self->verbose;
			# $self->{docobj} = $result[0] if $tag eq 'Doc';
		} else {
			$this = delete $args{caller} if $args{caller};
			# $this ||= $self->{docobj};
			$this ||= $self->var('$Doc');
			die "missing 'caller' for $tag" unless $this;
			print "$this->$tag(@{[%args]})" if $self->verbose;
			@result = $this->$tag(\%args);
			print " => @result\n" if $self->verbose;
		}
		$self->var($varname, @result) if $varname;
		if( $call ) {
			for my $obj(@result) {
				for my $tree(@$call) {
					$self->method_parse($tree, $obj);
				}
			}
		}
	}
	}; # end of eval
	die "error in '$tag': $@" if $@;
	@result;
}

sub args {
	my($self, $attributes, $children, $nosubst) = @_;
	my %args;
	%args = %$attributes if ref($attributes) eq 'HASH';
	%args = $self->arg_parse($children, %args) if $children;
	$self->subst_vars(\%args, $nosubst);
	if( ref($args{attributes}) eq 'HASH' ) {
		my $tmp = delete $args{attributes};
		$self->subst_vars($tmp, $nosubst);
		%args = (%args, %$tmp);
	}
	%args;
}

sub arg_parse {
	my($self, $tree, %args) = @_;
	return %args unless ref($tree) eq 'ARRAY';
	for my $elem($self->elements($tree)) {
		my($tag, $attributes, $children) = @$elem;
		next if $tag eq '0';
		if( $tag eq 'call' ) {
			push @{$args{$tag}}, $children;
		} else {
			for my $elem($self->elements($children)) {
				my($ctag, $cattr, $cchildren) = @$elem;
				my @value;
				if( $ctag eq '0' ) {
					@value = ($cchildren);
				} else {
					@value = $self->call_method($ctag, $cattr, $cchildren);
				}
				next unless @value;
				if( !exists $args{$tag} ) {
					$args{$tag} = @value == 1 ? $value[0] : \@value;
				} elsif( ref($args{$tag}) eq 'ARRAY' ) {
					push @{$args{$tag}}, @value;
				} else {
					$args{$tag} = [$args{$tag}, @value];
				}
			}
		}
	}
	%args;
}

sub subst_var {
	my($self, $str) = @_;
	$str = $self->var($str) if $str =~ /$VarnamePtn/;
	$str;
}

sub subst_vars {
	my($self, $args, $nosubst) = @_;
	return if !ref($nosubst) && $nosubst;
	my @nosubst = ref($nosubst) eq 'ARRAY' ? @$nosubst : ();
	my $nonosubst = 0;
	if( $nosubst[0] eq '!' ) {
		shift @nosubst;
		$nonosubst = 1;
	}
	for my $key(keys %$args) {
		next if $key eq 'setvar' || 
			($nonosubst ? !grep {$_ eq $key} @nosubst : 
			grep {$_ eq $key} @nosubst);
		if( ref($args->{$key}) eq 'ARRAY' ) {
			grep {$_ = $self->var($_) if /$VarnamePtn/ } @{$args->{$key}};
		} elsif( $args->{$key} =~ /$VarnamePtn/ ) {
			$args->{$key} = $self->var($args->{$key});
		}
	}
}

sub package {
	my($self) = @_;
	$self->{safe}->root;
}

sub var {
	my($self, $name, @values) = @_;
	$self->_var(0, $name, @values);
}

sub var_local {
	my($self, $name, @values) = @_;
	$self->_var(1, $name, @values);
}

sub _var {
	my($self, $local, $name, @values) = @_;
	my $pkg = $self->package;
	my $key;
	($name, $key) = $name =~ /^\$?([A-Za-z_]\w*)(?:\{([A-Za-z_]\w*)\})?$/;
	my $result;
	no strict 'refs';
	if( $key eq '' ) {
		if( $local ) {
			print "local($self->{locallevel}) $name => ",${$pkg.'::'.$name},"\n"
				if $self->verbose > 1;
			$self->{local}[$self->{locallevel}]{$name} = ${$pkg.'::'.$name};
		}
		if( @values ) {
			my $value = @values == 1 ? $values[0] : \@values;
			${$pkg.'::'.$name} = $value;
		}
		die "\$$name not defined" unless defined ${$pkg.'::'.$name};
		$result = ${$pkg.'::'.$name};
	} else {
		if( $local ) {
			print "local($self->{locallevel}) $name\{$key} => ",${$pkg.'::'.$name},"\n"
				if $self->verbose > 1;
			$self->{local}[$self->{locallevel}]{"$name\{$key}"} = ${$pkg.'::'.$name}{$key};
		}
		if( @values ) {
			my $value = @values == 1 ? $values[0] : \@values;
			${$pkg.'::'.$name}{$key} = $value;
		}
		die "\$$name\{$key} not defined" unless defined ${$pkg.'::'.$name}{$key};
		$result = ${$pkg.'::'.$name}{$key};
	}
	$result;
}

sub set_args {
	my($self, $args, $clear) = @_;
	my $pkg = $self->package;
	no strict 'refs';
	if( $clear ) {
		%{$pkg.'::Args'} = ();
	}
	if( ref($args) eq 'HASH' ) {
		for my $attr(keys %$args) {
			print "  $attr => $args->{$attr}\n" if $self->verbose > 1;
			if( $attr =~ s/^!// ) {
				delete ${$pkg.'::Args'}{$attr};
			} else {
				${$pkg.'::Args'}{$attr} = $args->{$attr};
			}
		}
	}
}

sub enter_local {
	my($self, $args) = @_;
	print "enter($self->{locallevel}) -->\n" if $self->verbose > 1;
	my $pkg = $self->package;
	no strict 'refs';
	my $tab = \%{$pkg.'::Args'};
	my $store = $self->{args}[$self->{locallevel}] = {};
	for my $key(keys %$tab) {
		$store->{$key} = $tab->{$key}; 
	}
	$self->set_args($args);
	$self->{locallevel}++;
	$self->{local}[$self->{locallevel}] = {};
}

sub leave_local {
	my($self) = @_;
	print "<-- leave($self->{locallevel})\n"  if $self->verbose > 1;
	die "locallevel must > 0" if $self->{locallevel} <= 0;
	my $pkg = $self->package;
	my $locals = $self->{local}[$self->{locallevel}];
	if( ref($locals) eq 'HASH' ) {
		no strict 'refs';
		for my $name(keys %$locals) {
			print "  $name => $locals->{$name}\n" if $self->verbose > 1;
			$self->var($name, $locals->{$name});
		}
	}
	$self->{locallevel}--;
	$self->set_args($self->{args}[$self->{locallevel}], 1);
}

sub eval {
	my($self, $code) = @_;
	$code = 'BEGIN{&$bytesimport()} ' . $code if $] > 5.007;
	$self->{safe}->reval($code);
}

sub split_call {
	my($self, $tree) = @_;
	return unless ref($tree) eq 'ARRAY';
	my(@call, @tmptree);
	for( my $j = 0; $j < @$tree - 1; $j += 2 ) {
		my($tag, $children) = ($tree->[$j], $tree->[$j + 1]);
		next if $tag eq '0' && $children =~ /^\s+$/ && $children =~ /[\r\n]/;
		if( $tag eq 'call' ) {
			push @call, $tag, $children;
		} else {
			push @tmptree, $tag, $children;
		}
	}
	(\@call, \@tmptree);
}

sub elements {
	my($self, $tree) = @_;
	#_elements($tree, $self->{alias});
	_elements($tree);
}

sub _elements {
	my($tree, $alias) = @_;
	my @elements;
	for( my $j = 0; $j < @$tree - 1; $j += 2 ) {
		my($tag, $children) = ($tree->[$j], $tree->[$j + 1]);
		next if $tag eq '0' && $children =~ /^\s+$/ && $children =~ /[\r\n]/;
		$tag = $alias->{$tag} if $alias && $alias->{$tag};
		my $attributes;
		if( ref($children) eq 'ARRAY' ) {
			my @children = @$children;
			$attributes = shift @children;
			my %tmp = %$attributes;
			$attributes = \%tmp;
			$children = \@children;
		}
		push @elements, [$tag, $attributes, $children];
	}
	@elements;
}

#---------------------
# internal method
#---------------------

sub do_require {
	my($self, $attributes, $children) = @_;
	my %args = $self->args($attributes, $children);
	die "version '$args{XPDFJ}' required but my version is '$VERSION'"
		if $args{XPDFJ} && $args{XPDFJ} > $VERSION;
	return;
}

sub do_eval {
	my($self, $attributes, $children) = @_;
	$self->_do_eval(0, $attributes, $children);
}

sub do_reval {
	my($self, $attributes, $children) = @_;
	$self->_do_eval(1, $attributes, $children);
}

sub _do_eval {
	my($self, $return, $attributes, $children) = @_;
	my $code = $self->method_parse($children, undef, undef, 1);
	my @result = $self->eval($code);
	die "eval error: $@: '$code'" if $@;
	return unless $return;
	@result;
}

sub do_for {
	my($self, $attributes, $children, $this) = @_;
	my %args = $self->args($attributes, undef, [qw(eval)]);
	my $varname = delete $args{setvar};
	my $eval = delete $args{eval};
	my @list = $self->eval($eval);
	@list = @{$list[0]} if @list == 1 && ref($list[0]) eq 'ARRAY';
	for my $value(@list) {
		$self->var($varname, $value);
		$self->method_parse($children, $this, \%args);
	}
	return;
}

sub do_def {
	my($self, $attributes, $children) = @_;
	my %args = $self->args($attributes, undef, 
		[qw(! tag attributesname contentsname)]);
	my $tag = delete $args{tag};
	my $attributesname = delete $args{attributesname};
	my $contentsname = delete $args{contentsname};
	my $contentsmode = delete $args{contentsmode};
	$self->{def}{$tag} = $children;
	$self->{attributesname}{$tag} = $attributesname if $attributesname;
	$self->{contentsname}{$tag} = $contentsname if $contentsname;
	$self->{contentsmode}{$tag} = $contentsmode if $contentsmode;
	$self->{defargs}{$tag} = \%args if %args;
	return;
}

sub do_alias {
	my($self, $attributes, $children) = @_;
	my %args = $self->args($attributes, $children, [qw(! tag aliasof)]);
	my $tag = delete $args{tag};
	my $aliasof = delete $args{aliasof};
	$self->{alias}{$tag} = $aliasof;
	$self->{aliasargs}{$tag} = \%args;
	return;
}

sub do_local {
	my($self, $attributes, $children) = @_;
	my %args = $self->args($attributes, undef, [qw(eval)]);
	my $varname = delete $args{setvar};
	my $eval = delete $args{eval};
	my @list = $eval ? $self->eval($eval) : ();
	$self->var_local($varname, @list);
	return;
}

sub do_do {
	my($self, $attributes, $children, $this) = @_;
	my %args = $self->args($attributes, undef, [qw(if unless result)]);
	my $if = delete $args{if};
	if( defined $if ) {
		return unless $self->eval($if);
	}
	my $unless = delete $args{unless};
	if( defined $unless ) {
		return if $self->eval($unless);
	}
	$this = delete $args{caller} if $args{caller};
	my $varname = delete $args{setvar};
	my $result = delete $args{result};
	my $file = delete $args{file};
	my $contents = delete $args{contents};
	my $withtext = delete $args{withtext};
	my $local = delete $args{local};
	my $verbose = delete $args{verbose};
	my $args = %args ? \%args : $local ? {} : undef;
	if( $file ) {
		my @path = split(/\s;\s/, $self->var('$XPDFJ{dopath}'));
		unless( -e $file ) {
			my $found = 0;
			for my $path( @path ) {
				if( -e "$path/$file" ) {
					$file = "$path/$file";
					$found = 1;
				}
			}
			die "missing file '$file'" unless $found;
		}
		print "do '$file'\n" if $self->verbose >= 0;
		my $parser = new XML::Parser(Style => 'Tree');
		$children = $parser->parsefile($file);
	} elsif( $contents ) {
		$children = $contents;
	}
	my $oldverbose;
	if( defined $verbose ) {
		$oldverbose = $self->verbose + 0;
		$self->verbose($verbose);
	}
	my @result = $self->method_parse($children, $this, $args, $withtext); 
	if( defined $verbose ) {
		$self->verbose($oldverbose);
	}
	$self->var($varname, @result) if $varname;
	if( $result eq 'first' ) {
		@result = shift @result;
	} elsif( $result eq 'last' ) {
		@result = pop @result;
	} elsif( $result eq 'null' ) {
		@result = ();
	} elsif( $result =~ /$VarnamePtn/ ) {
		@result = ($self->var($result));
	}
	@result;
}

sub do_sub {
	my($self, $attributes, $children, $this) = @_;
	return sub { $self->method_parse($children, $this); };
}

sub do_debuginfo {
	my($self, $attributes, $children, $this) = @_;
	my %args = $self->args($attributes);
	my $pattern = delete $args{pattern};
	print "locallevel: $self->{locallevel}\n";
	my $pkg = $self->package;
	no strict 'refs';
	my $tab = \%{$pkg.'::'};
	for my $name(keys %$tab) {
		next if $pattern && $name !~ /$pattern/;
		if( defined ${$tab->{$name}} ) { 
			print "  $name => ${$tab->{$name}}\n";
		}
		if( defined %{$tab->{$name}} && $name !~ /::$/ ) { 
			print "  $name => {\n";
			for my $key(keys %{$tab->{$name}}) {
				print "    $key => ",${$tab->{$name}}{$key},"\n";
			}
			print "  }\n";
		}
	}
	return;
}

# utility subroutines for eval package

sub istext {
	my($tree) = @_;
	return 1 unless ref($tree);
	return 0 unless ref($tree) eq 'ARRAY';
	$tree->[0] eq '0';
}

sub addcomma {
	my($num) = @_;
	$num += 0;
	while( $num =~ s/(\d)(\d\d\d)(?!\d)/$1,$2/g ) {}
	$num;
}

sub autonl {
	my($str) = @_;
	my @result;
	for my $part(split /[ \t]*\n[ \t]*/, $str) {
		next if $part eq '';
		push @result, $part, NewLine;
	}
	pop @result;
	@result;
}

1;

